import unittest

from n64img.image import (
    CI4,
    I4,
    IA4,
)


class TestEndian(unittest.TestCase):
    def create_4bpp_image(self, t, w, h):
        data = bytearray((w * h) >> 1)
        for i in range(0, len(data)):
            data[i] = i & 0xFF

        return t(data, w, h)

    def test_ci4_endian_big(self):
        img = self.create_4bpp_image(CI4, 4, 4)
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x03,
            0x00, 0x04, 0x00, 0x05, 0x00, 0x06, 0x00, 0x07,
        ]), pixels)

    def test_ci4_endian_little(self):
        img = self.create_4bpp_image(CI4, 4, 4)
        img.little_endian = True
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x03, 0x00,
            0x04, 0x00, 0x05, 0x00, 0x06, 0x00, 0x07, 0x00,
        ]), pixels)

    def test_i4_endian_big(self):
        img = self.create_4bpp_image(I4, 4, 4)
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x00, 0x11, 0x00, 0x22, 0x00, 0x33,
            0x00, 0x44, 0x00, 0x55, 0x00, 0x66, 0x00, 0x77,
        ]), pixels)

    def test_i4_endian_little(self):
        img = self.create_4bpp_image(I4, 4, 4)
        img.little_endian = True
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x11, 0x00, 0x22, 0x00, 0x33, 0x00,
            0x44, 0x00, 0x55, 0x00, 0x66, 0x00, 0x77, 0x00,
        ]), pixels)

    def test_ia4_endian_big(self):
        img = self.create_4bpp_image(IA4, 4, 4)
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0xFF,
            0x00, 0x00, 0x24, 0x00,
            0x00, 0x00, 0x24, 0xFF,
            0x00, 0x00, 0x49, 0x00,
            0x00, 0x00, 0x49, 0xFF,
            0x00, 0x00, 0x6D, 0x00,
            0x00, 0x00, 0x6D, 0xFF,
        ]), pixels)

    def test_ia4_endian_little(self):
        img = self.create_4bpp_image(IA4, 4, 4)
        img.little_endian = True
        pixels = img.parse()
        self.assertEqual(bytearray([
            0x00, 0x00, 0x00, 0x00,
            0x00, 0xFF, 0x00, 0x00,
            0x24, 0x00, 0x00, 0x00,
            0x24, 0xFF, 0x00, 0x00,
            0x49, 0x00, 0x00, 0x00,
            0x49, 0xFF, 0x00, 0x00,
            0x6D, 0x00, 0x00, 0x00,
            0x6D, 0xFF, 0x00, 0x00,
        ]), pixels)



import unittest

from n64img.image import (
    Image,
    CI4,
    CI8,
    I1,
    I4,
    I8,
    IA4,
    IA8,
    IA16,
    RGBA16,
    RGBA32,
)


class TestImage(unittest.TestCase):
    def test_size(self):
        img = Image(bytes(), 32, 32)
        self.assertEqual(1024, img.size())

    def test_mipmap_size(self):
        img = Image(bytes(), 32, 32)
        # 4x4 needs 4 pixels extra per row; (4+4)*1 = 8
        # 2x2 needs 6 pixels extra per row; (2+6)*1 = 8
        # 1x1 needs 7 pixels extra per row; (1+7)*1 = 8
        # 32*32*1 + 16*16*1 + 8*8*1 + 8*4*1 + 8*2*1 + 8*1*1
        self.assertEqual(1400, img.mipmap_size())


class TestCI4(unittest.TestCase):
    def test_size(self):
        img = CI4(bytes(), 32, 32)
        self.assertEqual(512, img.size())


class TestCI8(unittest.TestCase):
    def test_size(self):
        img = CI8(bytes(), 32, 32)
        self.assertEqual(1024, img.size())


class TestI1(unittest.TestCase):
    def test_size(self):
        img = I1(bytes(), 32, 32)
        self.assertEqual(128, img.size())


class TestI4(unittest.TestCase):
    def test_size(self):
        img = I4(bytes(), 32, 32)
        self.assertEqual(512, img.size())


class TestI8(unittest.TestCase):
    def test_size(self):
        img = I8(bytes(), 32, 32)
        self.assertEqual(1024, img.size())


class TestIA4(unittest.TestCase):
    def test_size(self):
        img = IA4(bytes(), 32, 32)
        self.assertEqual(512, img.size())


class TestIA8(unittest.TestCase):
    def test_size(self):
        img = IA8(bytes(), 32, 32)
        self.assertEqual(1024, img.size())


class TestIA16(unittest.TestCase):
    def test_size(self):
        img = IA16(bytes(), 32, 32)
        self.assertEqual(2048, img.size())


class TestRGBA16(unittest.TestCase):
    def test_size(self):
        img = RGBA16(bytes(), 32, 32)
        self.assertEqual(2048, img.size())

    def test_mipmap_size(self):
        img = RGBA16(bytes(), 32, 32)
        # 2x2 needs 2 pixels extra per row; (2+2)*2 = 8
        # 1x1 needs 3 pixels extra per row; (1+3)*2 = 8
        # 32*32*2 + 16*16*2 + 8*8*2 + 4*4*2 + 4*2*2 + 4*1*2
        self.assertEqual(2744, img.mipmap_size())


class TestRGBA32(unittest.TestCase):
    def test_size(self):
        img = RGBA32(bytes(), 32, 32)
        self.assertEqual(4096, img.size())

    def test_mipmap_size(self):
        img = RGBA32(bytes(), 32, 32)
        # 1x1 needs 1 pixels extra per row; (1+1)*4 = 8
        # 32*32*4 + 16*16*4 + 8*8*4 + 4*4*4 + 2*2*4 + 2*1*4
        self.assertEqual(5464, img.mipmap_size())

    def test_mipmap_size2(self):
        # TODO: sanity check this
        img = RGBA32(bytes(), 31, 31)
        # 31x31 needs 1 pixel extra per row; (31+1)*4 = 128
        # 15x15 needs 1 pixel extra per row; (15+1)*4 = 64
        # 7x7 needs 1 pixel extra per row; (7+1)*4 = 32
        # 3x3 needs 1 pixel extra per row; (3+1)*4 = 16
        # 1x1 needs 1 pixel extra per row; (1+1)*4 = 8
        # 32*31*4 + 16*15*4 + 8*7*4 + 4*3*4 + 2*1*4
        self.assertEqual(5208, img.mipmap_size())



from math import ceil
from pathlib import Path
from typing import List, Literal, Optional, Tuple

import png

from n64img import iter

Palette = List[Tuple[int, int, int, int]]


class Image:
    def __init__(self, data: bytes, width: int, height: int) -> None:
        self.data: bytes = data
        self.width: int = width
        self.height: int = height
        self.depth: float = 1
        self.greyscale: bool = False
        self.alpha: bool = False
        self.flip_h: bool = False
        self.flip_v: bool = False
        self.little_endian: bool = False
        self.palette: Optional[Palette] = None

    def __str__(self) -> str:
        return "Image(width={}, height={}, data={!r})".format(
            self.width, self.height, self.data
        )

    def set_palette(
        self, palette_bytes: bytes, endian: Literal["little", "big"] = "big"
    ) -> None:
        def unpack_color(data: bytes) -> Tuple[int, int, int, int]:
            s = int.from_bytes(data, endian)

            r = (s >> 11) & 0x1F
            g = (s >> 6) & 0x1F
            b = (s >> 1) & 0x1F
            a = (s & 1) * 0xFF

            r = ceil(0xFF * (r / 31))
            g = ceil(0xFF * (g / 31))
            b = ceil(0xFF * (b / 31))

            return r, g, b, a

        palette = []

        assert len(palette_bytes) % 2 == 0
        for i in range(0, len(palette_bytes), 2):
            palette.append(unpack_color(palette_bytes[i : i + 2]))

        self.palette = palette

    def get_writer(self) -> png.Writer:
        return png.Writer(
            self.width,
            self.height,
            greyscale=self.greyscale,
            alpha=self.alpha,
            palette=self.palette,
        )

    def parse(self) -> bytes:
        if not self.flip_h and not self.flip_v:
            return self.data

        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, self.depth, self.flip_h, self.flip_v
        ):
            img.append(self.data[i])

        return bytes(img)

    def write(self, outpath: Path) -> None:
        with open(outpath, "wb") as f:
            pixels = self.parse()
            self.get_writer().write_array(f, pixels)

    def mipmap_size(self) -> int:
        size = 0
        width = self.width
        while width > 0:
            # NOTE: rows must be padded to 8-byte boundary
            row = height = width
            remainder = ceil(width * self.depth) % 8
            if remainder != 0:
                row += ceil((8 - remainder) / self.depth)
            size += int(row * height * self.depth)
            width >>= 1
        return size

    def size(self) -> int:
        return ceil(self.width * self.height * self.depth)


class CI4(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 0.5

    def parse(self) -> bytes:
        img = bytearray()

        if self.little_endian:
            for x, y, i in iter.iter_image_indexes(
                self.width, self.height, self.depth, self.flip_h, self.flip_v
            ):
                img.append(self.data[i] & 0xF)
                img.append(self.data[i] >> 4)
        else:
            for x, y, i in iter.iter_image_indexes(
                self.width, self.height, self.depth, self.flip_h, self.flip_v
            ):
                img.append(self.data[i] >> 4)
                img.append(self.data[i] & 0xF)

        return bytes(img)


class CI8(Image):
    pass


# I1, a very primitive type where each bit represents one pixel.
# Generally, only used for debug fonts and rendered using the cpu instead of the rdp.
class I1(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 0.125
        self.greyscale = True

    def parse(self) -> bytes:
        if self.flip_h:
            raise Exception("I1 images cannot be flipped horizontally.")
        if self.flip_v:
            raise Exception("I1 images cannot be flipped vertically.")

        img = bytearray()

        for b in self.data:
            # iterate over bits
            for j in range(8, 0, -1):
                # Store the value of each bit as a pixel.
                p = (b >> (j - 1)) & 0x1
                # Convert active bits to RGB white and inactive to black.
                p *= 0xFF

                img.append(p)

        return bytes(img)


class I4(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 0.5
        self.greyscale = True

    def parse(self) -> bytes:
        img = bytearray()

        if self.little_endian:
            for x, y, i in iter.iter_image_indexes(
                self.width, self.height, self.depth, self.flip_h, self.flip_v
            ):
                b = self.data[i]

                i1 = b & 0xF
                i2 = (b >> 4) & 0xF

                i1 = (i1 << 4) | i1
                i2 = (i2 << 4) | i2

                img += bytes((i1, i2))
        else:
            for x, y, i in iter.iter_image_indexes(
                self.width, self.height, self.depth, self.flip_h, self.flip_v
            ):
                b = self.data[i]

                i1 = (b >> 4) & 0xF
                i2 = b & 0xF

                i1 = (i1 << 4) | i1
                i2 = (i2 << 4) | i2

                img += bytes((i1, i2))

        return bytes(img)


class I8(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.greyscale = True


class IA4(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 0.5
        self.greyscale = True
        self.alpha = True

    def parse(self) -> bytes:
        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, self.depth, flip_h=self.flip_h, flip_v=self.flip_v
        ):
            b = self.data[i]

            if self.little_endian:
                h = b & 0xF
                l = (b >> 4) & 0xF
            else:
                h = (b >> 4) & 0xF
                l = b & 0xF

            i1 = h & 0xE
            i1 = (i1 << 4) | (i1 << 1) | (i1 >> 2)
            a1 = (h & 1) * 0xFF

            i2 = l & 0xE
            i2 = (i2 << 4) | (i2 << 1) | (i2 >> 2)
            a2 = (l & 1) * 0xFF

            img += bytes((i1, a1, i2, a2))

        return bytes(img)


class IA8(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.greyscale = True
        self.alpha = True

    def parse(self) -> bytes:
        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, flip_h=self.flip_h, flip_v=self.flip_v
        ):
            b = self.data[i]

            i = (b >> 4) & 0xF
            a = b & 0xF

            i = (i << 4) | i
            a = (a << 4) | a

            img += bytes((i, a))

        return bytes(img)


class IA16(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 2
        self.greyscale = True
        self.alpha = True

    def parse(self) -> bytes:
        if not self.flip_h and not self.flip_v:
            return self.data

        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, self.depth, self.flip_h, self.flip_v
        ):
            img += bytes((self.data[i], self.data[i + 1]))

        return bytes(img)


class RGBA16(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 2
        self.greyscale = False
        self.alpha = True

    def parse(self) -> bytes:
        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, self.depth, self.flip_h, self.flip_v
        ):
            s = int.from_bytes(self.data[i : i + 2], byteorder="big")

            r = (s >> 11) & 0x1F
            g = (s >> 6) & 0x1F
            b = (s >> 1) & 0x1F

            r = (r << 3) | (r >> 2)
            g = (g << 3) | (g >> 2)
            b = (b << 3) | (b >> 2)

            a = 255 * (s & 1)

            img += bytes((r, g, b, a))

        return bytes(img)


class RGBA32(Image):
    def __init__(self, data: bytes, width: int, height: int) -> None:
        super().__init__(data, width, height)
        self.depth = 4
        self.greyscale = False
        self.alpha = True

    def parse(self) -> bytes:
        if not self.flip_h and not self.flip_v:
            return self.data

        img = bytearray()

        for x, y, i in iter.iter_image_indexes(
            self.width, self.height, self.depth, self.flip_h, self.flip_v
        ):
            img += bytes(
                (self.data[i], self.data[i + 1], self.data[i + 2], self.data[i + 3])
            )

        return bytes(img)



from itertools import zip_longest
from math import ceil
from typing import Iterable, Tuple


def iter_image_indexes(
    width: int,
    height: int,
    bytes_per_pixel: float = 1,
    flip_h: bool = False,
    flip_v: bool = False,
) -> Iterable[Tuple[int, int, int]]:
    w = int(width * bytes_per_pixel)
    h = int(height * 1)

    xrange = (
        range(w - ceil(bytes_per_pixel), -1, -ceil(bytes_per_pixel))
        if flip_h
        else range(0, w, ceil(bytes_per_pixel))
    )
    yrange = range(h - 1, -1, -1) if flip_v else range(0, h, 1)

    for y in yrange:
        for x in xrange:
            yield x, y, (y * w) + x
